import flask

from flaskel import httpcode, webargs
from flaskel.ext import auth, builder
from flaskel.ext.sqlalchemy import db
from . import bp_auth


class BlackListToken(db.Model, auth.RevokedTokenMixin):
    __tablename__ = "revoked_tokens"


handler = auth.DBTokenHandler(BlackListToken, db.session)


@bp_auth.route("/token/access", methods=["POST"])
@webargs.query(
    dict(
        expires_access=webargs.OptField.positive(),
        expires_refresh=webargs.OptField.positive(),
    )
)
def access_token(args):
    data = flask.request.json
    if data.email == "email" and data.password == "password":
        return handler.create(data.email, **args)
    flask.abort(httpcode.UNAUTHORIZED)


@bp_auth.route("/token/refresh", methods=["POST"])
@auth.jwt.jwt_required(refresh=True)
def refresh_token():
    return handler.refresh()


@bp_auth.route("/token/check", methods=["GET"])
@auth.jwt.jwt_required()
def check_token():
    return handler.dump()


@bp_auth.route("/token/revoke", methods=["POST"])
@builder.no_content
def revoke_token():
    data = flask.request.json
    if data.access_token:
        handler.revoke(data.access_token)
    if data.refresh_token:
        handler.revoke(data.refresh_token)



import os
import sys
import typing as t

import flask
import jinja2
from werkzeug.middleware.lint import LintMiddleware
from werkzeug.middleware.profiler import ProfilerMiddleware

from flaskel import config
from flaskel import flaskel
from .converters import CONVERTERS
from .utils import ObjectDict
from .utils import misc


class AppBuilder:  # pylint: disable=E1101
    """
    default app name
    """

    app_name = config.FLASK_APP

    """
    default flask class
    """
    flask_class = flaskel.Flaskel

    """
    default secret key file name
    """
    secret_file = ".secret.key"

    """
    custom url converters
    """
    url_converters = CONVERTERS

    """
    custom app config module
    """
    conf_module = config

    def __init__(
        self,
        conf_module=None,
        extensions=None,
        converters=None,
        blueprints=None,
        folders=None,
        middlewares=None,
        views=None,
        after_request=None,
        before_request=None,
        callback=None,
        version=None,
        **options,
    ):  # pylint: disable=too-many-arguments
        """

        :param conf_module: python module file
        :param converters: custom url converter mapping
        :param extensions: custom extension appended to defaults
        :param blueprints: list of application's blueprints
        :param folders: list of custom jinja2's template folders
        :param middlewares: list of wsgi middleware
        :param views: list of views to register
        :param after_request: list of functions to register after request
        :param before_request: list of functions to register before request
        :param callback: a function called to patch app after all components registration
        :param options: passed to Flask class
        :param version: app version, may be used by extensions
        :return:
        """
        self._app: t.Optional[flask.Flask] = None
        self._version = version
        self._converters = converters or {}
        self._blueprints = blueprints or ()
        self._extensions = extensions or {}
        self._middlewares = middlewares or ()
        self._folders = folders or ()
        self._options = options or {}
        self._views = views or ()
        self._after_request = after_request or ()
        self._before_request = before_request or ()
        self._callback = callback or (lambda: None)
        self._conf_module = conf_module or self.conf_module

    def _generate_secret_key(self, secret_file, key_length):
        secret_key = misc.random_string(key_length)

        with open(secret_file, "w") as f:
            f.write(secret_key)
            abs_file = os.path.abspath(secret_file)
            self._app.logger.warning(
                "new secret key generated: take care of this file: %s", abs_file
            )
        return secret_key

    def _load_secret_key(self, secret_file):
        if not os.path.isfile(secret_file):
            return None

        with open(secret_file, "r") as f:
            secret_key = f.read()
            self._app.logger.info("load secret key from: %s", secret_file)
        return secret_key

    def _set_secret_key(self):
        secret_key = None
        key_length = self._app.config["SECRET_KEY_MIN_LENGTH"]

        if not self._app.config.get("FLASK_ENV", "").lower().startswith("dev"):
            secret_file = self._app.config["SECRET_KEY"]
            if secret_file:
                secret_key = self._load_secret_key(secret_file) or secret_file
            else:
                secret_file = self.secret_file
                secret_key = self._load_secret_key(secret_file)
                secret_key = secret_key or self._generate_secret_key(
                    secret_file, key_length
                )
        elif not self._app.config.get("SECRET_KEY"):
            self._app.logger.debug("set secret key in development mode")
            secret_key = "fake_very_complex_string"

        self._app.config["SECRET_KEY"] = secret_key or self._app.config["SECRET_KEY"]
        if len(secret_key) < key_length:
            self._app.logger.warning("secret key length is less than: %s", key_length)

        if not self._app.config.get("JWT_SECRET_KEY") is None:
            self._app.config["JWT_SECRET_KEY"] = self._app.config["SECRET_KEY"]

    def _register_extensions(self):
        with self._app.app_context():
            for name, e in self._extensions.items():
                try:
                    ext, opt = e[0], e[1] if len(e) > 1 else {}
                    if not ext:
                        raise TypeError("extension could not be None or empty")
                except (TypeError, IndexError) as exc:
                    self._app.logger.warning(
                        "Invalid extension '%s' configuration '%s': %s", name, e, exc
                    )
                    continue

                ext.init_app(self._app, **opt)
                self._app.logger.debug(
                    "Registered extension '%s' with options: %s", name, opt
                )

        self._app.logger.debug("Dump flask extensions:")
        for k, v in self._app.extensions.items():
            self._app.logger.debug("Registered extension '%s': %s", k, v)

    def _register_blueprints(self):
        for b in self._blueprints:
            try:
                bp, opt = b[0], b[1] if len(b) > 1 else {}
                if not bp:
                    raise TypeError("blueprint could not be None or empty")
            except (TypeError, IndexError) as exc:
                self._app.logger.warning(
                    "invalid blueprint configuration '%s': %s", b, exc
                )
                continue

            self._app.register_blueprint(bp, **opt)
            self._app.logger.debug(
                "Registered blueprint '%s' with options: %s", bp.name, opt
            )

    def _set_config(self, conf):
        self._app.config.from_object(self._conf_module)
        self._app.config.from_mapping(**(conf or {}))
        self._app.config.from_envvar("APP_CONFIG_FILE", silent=True)
        self._app.config = ObjectDict(**self._app.config)

    def _register_converters(self):
        self._app.url_map.converters.update({**self.url_converters, **self._converters})
        for k, v in self._app.url_map.converters.items():
            self._app.logger.debug("Registered converter: '%s' = %s", k, v.__name__)

    def _register_template_folders(self):
        loaders = [self._app.jinja_loader]

        for fsl in self._folders:
            loaders.append(jinja2.FileSystemLoader(fsl))

        if self._folders:
            self._app.jinja_loader = jinja2.ChoiceLoader(loaders)
            for f in self._folders:
                self._app.logger.debug("Registered template folder: '%s'", f)

    def _register_middlewares(self):
        for m in self._middlewares:
            kwargs = {}
            try:
                if isinstance(m, (list, tuple)):
                    m, kwargs = m[0], m[1] if len(m) > 1 else {}
                if not m:
                    raise TypeError("middleware could not be None or empty")
            except (TypeError, IndexError) as exc:
                self._app.logger.warning(
                    "invalid middleware configuration '%s': %s", m, exc
                )
                continue

            # WorkAround: in order to pass flask app to middleware without breaking chain
            if not (hasattr(m, "flask_app") and m.flask_app):
                setattr(m, "flask_app", self._app)

            self._app.wsgi_app = m(self._app.wsgi_app, **kwargs)
            self._app.logger.debug("Registered middleware: '%s'", m.__name__)

    def _register_views(self):
        def normalize(data):
            d = data + (None,) * (3 - len(data))
            if isinstance(d[1], dict):
                return d[0], d[2], d[1]
            return d[0], d[1], d[2] or {}

        for view in self._views:
            v, b, p = normalize(view)
            v.register(b or self._app, **p)
            mess = f"on blueprint '{b.name}'" if b else ""
            self._app.logger.debug(
                "Registered view: '%s' %s with params: %s", v.__name__, mess, p
            )

    def _register_after_request(self):
        for f in self._after_request:
            self._app.after_request_funcs.setdefault(None, []).append(f)
            self._app.logger.debug(
                "Registered function after request: '%s'", f.__name__
            )

    def _register_before_request(self):
        for f in self._before_request:
            self._app.before_request_funcs.setdefault(None, []).append(f)
            self._app.logger.debug(
                "Registered function before request: '%s'", f.__name__
            )

    def _set_linter_and_profiler(self):
        if self._app.config["WSGI_WERKZEUG_LINT_ENABLED"]:
            self._app.wsgi_app = LintMiddleware(self._app.wsgi_app)
            self._app.logger.debug(
                "Registered middleware: '%s'", LintMiddleware.__name__
            )

        if self._app.config["WSGI_WERKZEUG_PROFILER_ENABLED"]:
            file = self._app.config["WSGI_WERKZEUG_PROFILER_FILE"]
            self._app.wsgi_app = ProfilerMiddleware(
                self._app.wsgi_app,
                stream=open(file, "w") if file else sys.stdout,  # pylint: disable=R1732
                restrictions=self._app.config["WSGI_WERKZEUG_PROFILER_RESTRICTION"],
            )
            self._app.logger.debug(
                "Registered middleware: '%s'", ProfilerMiddleware.__name__
            )

    def _dump_urls(self):
        class DumpUrls:
            def __init__(self, rules):
                self._rules = rules

            def __str__(self):
                output = []
                for rule in self._rules:
                    methods = ",".join(rule.methods)
                    output.append(
                        "{:30s} {:40s} {}".format(rule.endpoint, methods, rule)
                    )
                return "\n".join(sorted(output))

        self._app.logger.debug(
            "Registered routes:\n%s", DumpUrls(self._app.url_map.iter_rules())
        )

    def _patch_app(self):
        if self._app.debug:
            self._set_linter_and_profiler()
            self._dump_urls()

        with self._app.app_context():
            sqlalchemy = self._app.extensions.get("sqlalchemy")
            if sqlalchemy is not None:
                sqlalchemy.db.create_all()

            self._callback()

    def create(self, conf=None):
        self._app = self.flask_class(self.app_name, **self._options)
        self._app.version = self._version

        self._set_config(conf)
        self._set_secret_key()

        self._register_extensions()
        self._register_middlewares()
        self._register_template_folders()
        self._register_converters()
        self._register_views()
        self._register_blueprints()
        self._register_after_request()
        self._register_before_request()

        self._patch_app()

    def get_or_create(self, conf=None):
        if self._app is None:
            self.create(conf)

        return self._app



from flaskel.extra.mobile_support import MobileLoggerView, MobileReleaseView
from flaskel.views import rpc
from flaskel.extra import apidoc

bp_api = None  # import blueprint here
rpc_service = object  # import rpc_service here
rpc.JSONRPCView.load_from_object(rpc_service())

VIEWS = (
    (
        rpc.JSONRPCView,
        dict(name=rpc.JSONRPCView.default_view_name, url=rpc.JSONRPCView.default_url),
    ),
    (
        apidoc.ApiDocTemplate,
        bp_api,
        dict(
            name=apidoc.ApiDocTemplate.default_view_name,
            urls=apidoc.ApiDocTemplate.default_urls,
        ),
    ),
    (
        apidoc.ApiSpecTemplate,
        bp_api,
        dict(
            name=apidoc.ApiSpecTemplate.default_view_name,
            urls=apidoc.ApiSpecTemplate.default_urls,
        ),
    ),
    (
        MobileReleaseView,
        bp_api,
        dict(
            name=MobileReleaseView.default_view_name,
            urls=MobileReleaseView.default_urls,
        ),
    ),
    (
        MobileLoggerView,
        bp_api,
        dict(
            name=MobileLoggerView.default_view_name,
            urls=MobileLoggerView.default_urls,
        ),
    ),
)



try:
    from tornado.httpserver import HTTPServer
    from tornado.ioloop import IOLoop
    from tornado.wsgi import WSGIContainer
except ImportError:
    HTTPServer = IOLoop = WSGIContainer = object

from .base import BaseApplication


class WSGITornado(BaseApplication):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        assert HTTPServer is not object, "you must install 'tornado'"
        BaseApplication.__init__(self, app, options)
        self._http_server = HTTPServer(WSGIContainer(self.application))

    def run(self):
        self._http_server.listen(address=self._interface, port=self._port)
        IOLoop.current().start()



import enum

from flask import current_app as cap


class ConfigProxy:
    def __init__(self, key=None):
        self.key = key

    def __getattr__(self, item):
        obj = self._proxy() or {}
        if not isinstance(item, str):
            return obj.get(item)
        for i in item.split("."):
            obj = obj.get(i)
        return obj

    def __call__(self, item=None, **kwargs):
        if item is None:
            return self._proxy()
        return self.__getattr__(item)

    def __getitem__(self, item):
        if item is None:
            return self._proxy()
        return self.__getattr__(item)

    def _proxy(self):
        res = cap.config

        if self.key is None:
            return cap.config
        for c in self.key.split("."):
            res = res.get(c) or {}

        return res or None

    def get(self, item=None):
        return self.__call__(item)


class ExtProxy:
    def __init__(self, name):
        self._name = name

    @property
    def extension(self):
        attrs = self._name.split(".")
        ret = cap.extensions[attrs[0]]
        for a in attrs[1:]:
            ret = getattr(ret, a, None)

        return ret

    def __getattr__(self, item):
        return getattr(self.extension, item)

    def __call__(self, *args, **kwargs):
        return self.extension()

    def __getitem__(self, item):
        return self.extension[item]


class HashableDict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


class ObjectDict(dict):
    def __init__(self, **kwargs):
        super().__init__()
        for k, v in kwargs.items():
            self[k] = self.normalize(v)

    def __dict__(self):
        data = {}
        for k, v in self.items():
            if isinstance(v, ObjectDict):
                data[k] = v.__dict__()
            elif isinstance(v, list):
                data[k] = []
                for i in v:
                    data[k].append(i.__dict__() if hasattr(i, "__dict__") else i)
            else:
                data[k] = v

        return data

    def __getstate__(self):
        return self.__dict__()

    def __setstate__(self, state):
        for k, v in state.items():
            self.__setattr__(k, v)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        return None

    def __setattr__(self, name, value):
        self[name] = self.normalize(value)

    def __delattr__(self, name):
        if name in self:
            del self[name]

    def patch(self, m, **kwargs):
        """

        :param m:
        :param kwargs:
        :return:
        """
        super().update(m, **kwargs)
        return self

    @staticmethod
    def normalize(data):
        """

        :param data:
        :return:
        """
        try:
            if isinstance(data, (list, tuple, set)):
                return [ObjectDict(**r) if isinstance(r, dict) else r for r in data]
            return ObjectDict(**data)
        except (TypeError, ValueError, AttributeError):
            return data

    def get_namespace(self, prefix, lowercase=True, trim=True):
        """
        Returns a dictionary containing a subset of configuration options
        that match the specified prefix.

        :param prefix: a configuration prefix
        :param lowercase: a flag indicating if the keys should be lowercase
        :param trim: a flag indicating if the keys should include the namespace
        """
        res = {}
        for k, v in self.items():
            if k.startswith(prefix):
                key = k[len(prefix) :] if trim else k
                res[key.lower() if lowercase else key] = v
        return res


class IntEnum(enum.IntEnum):
    @classmethod
    def to_list(cls):
        return [
            ObjectDict(id=getattr(cls, m).value, label=getattr(cls, m).name)
            for m in cls.__members__
        ]

    def to_dict(self):
        return ObjectDict(id=self.value, label=self.name)

    def __repr__(self):
        return f"<{self.value}: {self.name}>"

    def __str__(self):
        return self.name


class Dumper:
    def __init__(self, data, *args, callback=None, **kwargs):
        """

        :param data:
        :param callback:
        :param args:
        :param kwargs:
        """
        self.data = data
        self._args = args
        self._kwargs = kwargs
        self._callback = callback

    def dump(self):
        if self._callback is not None:
            return self._callback(self.data, *self._args, **self._kwargs)
        return str(self.data)

    def __str__(self):
        return self.dump()



# noinspection PyUnresolvedReferences
from flaskel.config import *  # noqa: F403,F401,E0401 pylint: disable=unused-wildcard-import,wildcard-import

LOCALE = "it_IT.utf8"
ERROR_PAGE = "errors.html"
ERROR_HANDLER = "web"



from flaskel.utils.datetime import DateHelper


class FlaskDateHelper:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        :param kwargs:
        """
        self._helper = None
        self.iso_format = None

        if app is not None:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, helper=None, helper_class=DateHelper, **__):
        """

        :param app:
        :param helper:
        :param helper_class:
        """
        app.config.setdefault("DATE_HELPER_COUNTRY", "IT")
        app.config.setdefault("DATE_HELPER_PROV", None)
        app.config.setdefault("DATE_HELPER_STATE", None)
        app.config.setdefault("DATE_PRETTY", helper_class.DATE_PRETTY_FORMAT)
        app.config.setdefault("DATE_ISO_FORMAT", helper_class.DATE_ISO_FORMAT)

        if helper is not None:
            self._helper = helper
        else:
            attrs = app.config.get_namespace("DATE_HELPER_")
            self._helper = helper_class(**attrs)

        self.iso_format = app.config.DATE_ISO_FORMAT
        if not hasattr(app, "extensions"):
            app.extensions = {}  # pragma: no cover
        app.extensions["date_helper"] = self

    def __getattr__(self, name):
        return getattr(self._helper, name)

    def to_iso_format(self, str_date, fmt=None, **kwargs):
        return self.change_format(
            str_date, in_fmt=fmt, out_fmt=self.iso_format, **kwargs
        )


date_helper = FlaskDateHelper()



import typing as t

import flask

from .base import BaseView


class RenderTemplate(BaseView):
    methods = ["GET", "POST"]
    template: t.Union[str, None] = None

    def __init__(self, template=None, content_type=None, **kwargs):
        """

        :param template:
        :param content_type:
        """
        self._params = kwargs
        self._content_type = content_type
        self._template = template or self.template

    # noinspection PyUnusedLocal
    def service(self, *_, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return kwargs or self._params

    def response(self, data, **kwargs):
        return flask.Response(data, mimetype=self._content_type, **kwargs)

    def dispatch_request(self, *_, **__):
        params = self.service()
        template = flask.render_template(self._template, **params)
        return self.response(template)


class RenderTemplateString(RenderTemplate):
    def dispatch_request(self, *_, **__):
        params = self.service()
        template = flask.render_template_string(self._template, **params)
        return self.response(template)



from flask_response_builder.builders.encoders import JsonEncoder

try:
    import redis
except ImportError:  # pragma: no cover
    redis = None  # type: ignore

try:
    from rejson import Client, Path
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    Client = Path = object


class JSONRedisClient(Client):
    def __init__(self, *args, encoder=None, **kwargs):
        super().__init__(encoder or JsonEncoder(), *args, **kwargs)

    @staticmethod
    def json_path(path=None):
        return Path(path) if path else Path.rootPath()

    def json_set(self, key, data, path=None):
        return self.jsonset(key, self.json_path(path), data)

    def json_get(self, key, path=None):
        return self.jsonget(key, self.json_path(path))

    def json_array_add(self, key, data, path=None, **__):
        path = self.json_path(path)
        pipe = self.pipeline()
        pipe.jsonset(key, path, [], nx=True)
        pipe.jsonarrappend(key, path, data)
        return pipe.execute()[1]


class FlaskRedis:
    def __init__(self, app=None, client=None, **kwargs):
        """

        :param app:
        :param client:
        :param kwargs:
        """
        self._client = client

        if app is not None:
            self.init_app(app, **kwargs)  # pragma: no cover

    @property
    def client(self):
        return self._client  # pragma: no cover

    def init_app(self, app, redis_class=None, **kwargs):
        """

        :param app:
        :param redis_class:
        """
        assert redis is not None, "redis must be installed"
        if redis_class is None:
            if Client is object:
                redis_class = redis.StrictRedis
            else:
                redis_class = JSONRedisClient

        app.config.setdefault("REDIS_URL", "redis://localhost:6379/0")
        app.config.setdefault("REDIS_OPTS", **kwargs)

        if not self._client:
            self._client = redis_class.from_url(
                app.config["REDIS_URL"], **app.config["REDIS_OPTS"]
            )

        if not hasattr(app, "extensions"):
            app.extensions = {}  # pragma: no cover
        app.extensions["redis"] = self

    def __getattr__(self, name):
        return getattr(self._client, name)


client_redis = FlaskRedis()



import json
import logging

from flaskel import httpcode
from flaskel.http import FlaskelHTTPDumper
from flaskel.utils.schemas.default import SCHEMAS
from .helpers import basic_auth_header
from .mixins import JSONValidatorMixin, RegexMixin

__all__ = [
    "TestHttpApi",
    "TestHttpCall",
    "TestJsonRPC",
    "TestRestApi",
]


class TestHttpCall(FlaskelHTTPDumper, JSONValidatorMixin, RegexMixin):
    __test__ = False
    default_status_code = (httpcode.SUCCESS, 299)
    default_content_type = "text/html"

    def __init__(self, test_client, endpoint=None, auth=None, **__):
        """

        :param endpoint:
        :param auth:
        :param kwargs:
        """
        self.auth = None
        self.response = None
        self.endpoint = endpoint
        self.test_client = test_client
        self.log = logging.getLogger()

        if auth:
            self.set_auth(auth)

    def set_url(self, url):
        """

        :param url:
        """
        if not url.startswith("http"):
            self.endpoint = "/".join([self.endpoint.rstrip("/"), url.lstrip("/")])
        else:
            self.endpoint = url

    def set_auth(self, config):
        """

        :param config:
        """
        config = config or {}
        auth_type = config.get("type")
        if auth_type == "basic" and config.get("username"):
            password = config.get("password") or config["username"]
            self.auth = basic_auth_header(config.get("username"), password)

    def assert_status_code(self, code, in_range=False, is_in=False):
        """

        :param code:
        :param in_range:
        :param is_in:
        """
        status_code = self.response.status_code
        if type(code) in (list, tuple):
            if in_range:
                self.assert_range(status_code, code)
            if is_in:
                self.assert_in(status_code, code)
        else:
            self.assert_equals(status_code, code)

    def assert_header(self, name, value, is_in=False, regex=False):
        """

        :param name:
        :param value:
        :param is_in:
        :param regex:
        """
        header = self.response.headers.get(name)
        if is_in:
            self.assert_in(header, value)
        elif regex:
            self.assert_match(header, value)
        else:
            self.assert_equals(header, value)

    def assert_response(self, **kwargs):
        """

        :param kwargs:
        """
        code = kwargs.get("status") or {
            "code": self.default_status_code,
            "in_range": True,
        }
        headers = kwargs.get("headers") or {}
        if "Content-Type" not in headers:
            headers["Content-Type"] = {
                "value": rf"{self.default_content_type}*",
                "regex": True,
            }

        self.assert_status_code(**code)
        for k, v in headers.items():
            if v is not None:
                self.assert_header(name=k, **v)

    def request(self, method="GET", url=None, **kwargs):
        """

        :param method:
        :param url:
        :param kwargs:
        :return:
        """
        url = url or self.endpoint
        if not url.startswith("http"):
            url = "{}{}".format(self.endpoint, url)

        self.set_auth(kwargs.pop("auth", None))
        if self.auth is not None:
            kwargs["auth"] = self.auth

        self.response = self.test_client.open(method=method, path=url, **kwargs)
        self.dump_response(self.response, dump_body=True)

    def perform(self, request, response=None, **__):
        """

        :param request:
        :param response:
        """
        self.request(**request)
        self.assert_response(**(response or {}))


class TestHttpApi(TestHttpCall):
    default_content_type = "application/json"

    @property
    def json(self):
        try:
            if self.response.status_code != httpcode.NO_CONTENT and "json" in (
                self.response.headers.get("Content-Type") or ""
            ):
                return self.response.get_json()
        except json.decoder.JSONDecodeError as exc:
            assert False, "Test that json is valid failed, got: {}".format(exc)
        return None

    def assert_response(self, **kwargs):
        """

        :param kwargs:
        """
        super().assert_response(**kwargs)
        if kwargs.get("schema") is not None:
            self.assert_schema(self.json, kwargs.get("schema"))


class TestJsonRPC(TestHttpApi):
    version = "2.0"
    default_schema = SCHEMAS.JSONRPC.response

    # noinspection PyShadowingBuiltins
    @classmethod
    def prepare_payload(
        cls, id=False, method=None, params=None
    ):  # pylint: disable=W0622
        """

        :param id:
        :param method:
        :param params:
        :return:
        """
        data = dict(jsonrpc=cls.version, method=method)
        if id is not False:
            data["id"] = id
        if params:
            data["params"] = params
        return data

    def perform(self, request, response=None, **__):
        """

        :param request:
        :param response:
        """
        req = {"method": "POST"}
        res = response or {}
        res.setdefault("schema", self.default_schema)

        if type(request) in (list, tuple):
            req["json"] = [self.prepare_payload(*a) for a in request]
        else:
            req["json"] = self.prepare_payload(**request)

        super().perform(request=req, response=res)


class TestRestApi(TestHttpApi):
    def __init__(self, *args, resource=None, res_id=None, **kwargs):
        """

        :param resource:
        :param res_id: resource id key
        """
        super().__init__(*args, **kwargs)
        self.res_id = res_id or "id"
        self.set_resource(resource)

    def set_resource(self, resource, res_id=None):
        """

        :param resource:
        :param res_id:
        """
        self.res_id = res_id or self.res_id
        if resource:
            self.endpoint = f"{self.endpoint}/{resource}"

    def resource_url(self, res_id):
        """

        :param res_id:
        :return:
        """
        return f"{self.endpoint}/{res_id}"

    def _normalize(self, request, response, method="GET", url=None):
        """

        :param request:
        :param response:
        :param method:
        :param url:
        :return:
        """
        request = request or {}
        response = response or {}
        request["method"] = method
        request["url"] = url or self.endpoint
        return request, response

    def test_collection(self, request=None, response=None):
        req, res = self._normalize(request, response)
        res.setdefault(
            "status",
            dict(code=(httpcode.SUCCESS, httpcode.PARTIAL_CONTENT), is_in=True),
        )
        self.perform(req, res)

    def test_post(self, request=None, response=None):
        req, res = self._normalize(request, response, "POST")
        res.setdefault("status", dict(code=httpcode.CREATED))
        self.perform(req, res)

    def test_get(self, res_id, request=None, response=None):
        req, res = self._normalize(request, response, url=self.resource_url(res_id))
        self.perform(req, res)

    def test_put(self, res_id, request=None, response=None):
        req, res = self._normalize(request, response, "PUT", self.resource_url(res_id))
        self.perform(req, res)

    def test_delete(self, res_id, request=None, response=None):
        req, res = self._normalize(
            request, response, "DELETE", self.resource_url(res_id)
        )
        res.setdefault("status", dict(code=httpcode.NO_CONTENT))
        res.setdefault("headers", {})
        res["headers"]["Content-Type"] = None
        self.perform(req, res)



# generated via cli
import os
import re
import sys

from setuptools import find_packages as base_find_packages
from setuptools import setup
from setuptools.command.test import test

LICENSE = "MIT"
URL = None
PLATFORMS = "any"
PYTHON_VERSION = ">3.6"
DESCRIPTION = None
PACKAGE_DATA = True

PKG_NAME = "{skeleton}"
PKG_TEST = "tests"
PKG_SCRIPTS = f"{PKG_NAME}.scripts"

EXCLUDE_FILES = []  # type: ignore

REQUIRES = ["flaskel"]

ENTRY_POINTS = dict(
    console_scripts=[
        f"run-{PKG_NAME}={PKG_SCRIPTS}.cli:cli",
    ],
)

BASE_PATH = os.path.dirname(__file__)
VERSION_FILE = os.path.join(PKG_NAME, "skel/version.py")

try:
    # must be after setuptools
    # noinspection PyPackageRequirements
    from Cython.Build import cythonize as base_cythonize

    # noinspection PyPackageRequirements
    import Cython.Compiler.Options as cython_options

    cython_options.docstrings = False
except ImportError:
    cython_options = None
    base_cythonize = None

if "--cythonize" not in sys.argv:
    base_cythonize = None
else:
    sys.argv.remove("--cythonize")


def ext_paths(root_dir, exclude=()):
    paths = []
    for root, dirs, files in os.walk(root_dir):
        for filename in files:
            file_path = os.path.join(root, filename)
            file_ext = filename.split(".")[-1]
            if file_path in exclude or file_ext not in ("py", "pyx"):
                continue

            paths.append(file_path)
    return paths


def read(file):
    with open(os.path.join(BASE_PATH, file)) as f:
        return f.read()


def grep(file, name):
    (value,) = re.findall(fr"{name}\W*=\W*'([^']+)'", read(file))
    return value


def readme(file):
    try:
        return read(file)
    except OSError as exc:
        print(str(exc), file=sys.stderr)


class PyTest(test):
    def finalize_options(self):
        test.finalize_options(self)

    def run_tests(self):
        import pytest

        sys.exit(pytest.main([PKG_TEST]))


def cythonize(paths):
    if base_cythonize is not None:
        return base_cythonize(paths, language_level=3)


def find_packages():
    if base_cythonize is not None:
        return [PKG_SCRIPTS]
    return base_find_packages(exclude=(PKG_TEST, f"{PKG_TEST}.*"))


setup(
    name=PKG_NAME,
    url=URL,
    license=LICENSE,
    description=DESCRIPTION,
    platforms=PLATFORMS,
    python_requires=PYTHON_VERSION,
    long_description=readme("README.rst"),
    version=grep(VERSION_FILE, "__version__"),
    author=grep(VERSION_FILE, "__author_name__"),
    author_email=grep(VERSION_FILE, "__author_email__"),
    zip_safe=False,
    include_package_data=PACKAGE_DATA,
    packages=find_packages(),
    ext_modules=cythonize(ext_paths(PKG_NAME, EXCLUDE_FILES)),
    entry_points=ENTRY_POINTS,
    test_suite=PKG_TEST,
    install_requires=REQUIRES,
    cmdclass=dict(test=PyTest),
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "License :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ],
)



import pytest

MARKERS = {
    "smoke": {"description": "smoke tests", "marker": pytest.mark.smoke},
    "email": {"description": "email tests", "marker": pytest.mark.email},
    "integ": {"description": "integration tests", "marker": pytest.mark.integ},
    "infra": {"description": "infrastructure tests", "marker": pytest.mark.infra},
}


def pytest_configure(config):
    """

    :param config:
    """
    for k, v in MARKERS.items():
        config.addinivalue_line("markers", "{}: {}".format(k, v["description"]))



import sys

import pytest


def main():
    pytest.main(
        [
            "-v",
            "-rf",
            "--strict-markers",
            "-p",
            "flaskel.tester.plugins.fixtures",
            "-p",
            "flaskel.tester.plugins.startup",
            *sys.argv[1:],
        ]
    )


if __name__ == "__main__":
    main()



import flask

from flaskel import cap, httpcode
from flaskel.http.client import FlaskelHttp
from flaskel.utils.datastruct import ObjectDict
from .base import BaseView


class ProxyView(BaseView):
    def __init__(
        self,
        host=None,
        url=None,
        method=None,
        proxy_body=False,
        proxy_headers=False,
        proxy_params=False,
        options=None,
        skip_args=(),
        **kwargs,
    ):
        """

        :param host:
        :param url:
        :param method:
        :param proxy_body:
        :param proxy_headers:
        :param proxy_params:
        :param skip_args: a tuple o arguments name to remove from kwargs
                         it is necessary because flask pass url params to dispatch_request
        :param options: callable that returns dict to pass to http client instance
                        It overrides kwargs
        """
        self._host = host
        self._method = method
        self._url = url
        self._proxy_body = proxy_body
        self._proxy_headers = proxy_headers
        self._proxy_params = proxy_params
        self._options = kwargs
        self._skip_args = skip_args

        if callable(options):
            self._options = {**kwargs, **options()}  # pragma: no cover

    def _filter_kwargs(self, data):
        for arg in self._skip_args:
            data.pop(arg, None)
        return data

    def dispatch_request(self, *_, **kwargs):
        """

        :param kwargs:
        :return:
        """
        opts = {**self._options, **self._filter_kwargs(kwargs)}
        response = self.proxy(self.service(), **opts)

        return flask.Response(
            flask.stream_with_context(response.body),
            headers=response.headers,
            status=response.status,
        )

    def proxy(self, data, **kwargs):
        client = FlaskelHttp(data.host or self.upstream_host(), **kwargs)
        return client.request(
            data.url or self.request_url(),
            method=data.method or self.request_method(),
            headers=data.headers or self.request_headers(),
            params=data.params or self.request_params(),
            data=data.body or self.request_body(),
            stream=True,
        )

    def service(self):
        """

        :return: ObjectDict
        """
        return ObjectDict(
            host=self.upstream_host(),
            uri=self.request_url(),
            method=self.request_method(),
            headers=self.request_headers(),
            params=self.request_params(),
            data=self.request_body(),
        )

    def upstream_host(self):
        return self._host

    def request_url(self):
        return self._url or flask.request.path

    def request_method(self):
        return self._method or flask.request.method

    def request_body(self):
        return flask.request.get_data() if self._proxy_body else None

    def request_headers(self):
        return flask.request.headers if self._proxy_headers else None

    def request_params(self):
        return flask.request.args if self._proxy_params else None


class ConfProxyView(BaseView):
    sep = "."
    config_key = None

    def __init__(self, config_key=None):
        self._config_key = config_key or config_key
        assert self._config_key is not None, "missing 'config_key'"

    def dispatch_request(self, *args, **kwargs):
        return self.perform(*args, **kwargs)

    def perform(self, *_, item=None, **__):
        """

        :param item:
        :return: ObjectDict
        """
        conf = cap.config
        keys = self._config_key.split(self.sep)
        if item is not None:
            keys += item.split(self.sep)
        for k in keys:
            conf = conf.get(k) or ObjectDict()
        if not conf:
            cap.logger.warning(f"unable to find conf keys: {self.sep.join(keys)}")
            flask.abort(httpcode.NOT_FOUND)
        return ObjectDict(**conf)


class TransparentProxyView(ProxyView):
    methods = [
        "POST",
        "PUT",
        "GET",
        "DELETE",
    ]

    def __init__(self, **kwargs):
        kwargs.setdefault("proxy_body", True)
        kwargs.setdefault("proxy_headers", True)
        kwargs.setdefault("proxy_params", True)
        super().__init__(**kwargs)


class SchemaProxyView(ConfProxyView):
    default_urls = [
        "/schema/<path:filepath>",
    ]

    def __init__(
        self,
        config_key="SCHEMAS",
        case_sensitive=False,
        ext_support=True,
        ext_name=".json",
    ):
        super().__init__(config_key)
        self.ext_name = ext_name
        self.ext_support = ext_support
        self.case_sensitive = case_sensitive

    def normalize(self, filepath):
        if self.ext_support and filepath.endswith(self.ext_name):
            filepath = self.sep.join(filepath.split(self.sep)[:-1])
        if self.case_sensitive is False:
            filepath = filepath.upper()

        return filepath.replace("/", self.sep)

    def dispatch_request(self, filepath, *args, **kwargs):
        schema_path = self.normalize(filepath)
        return super().perform(item=schema_path)



import functools

import flask

from flaskel.ext.default import builder
from flaskel.flaskel import cap
from flaskel.flaskel import httpcode
from flaskel.utils.batch import BatchExecutor
from flaskel.utils.batch import ThreadBatchExecutor


class HealthCheck:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        """
        self._executor = None
        self._health_checks = {}

        if app:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, bp=None, extensions=(), decorators=(), executor=None):
        """

        :param app: Flask app or blueprint instance
        :param bp: optional a blueprint instance on which register the healthcheck view
        :param extensions: functions and arguments that execute the checks
        :param decorators: functions that decorates the healthcheck view
        :param executor: batch executor class, must be subclass of BatchExecutor
        """
        self._executor = executor or ThreadBatchExecutor
        assert issubclass(self._executor, BatchExecutor)

        self.set_default_config(app)
        self.register_route(bp or app, decorators)
        self.register_extensions(app, extensions)

        if not hasattr(app, "extensions"):
            app.extensions = dict()  # pragma: no cover
        app.extensions["healthcheck"] = self

    @staticmethod
    def set_default_config(app):
        app.config.setdefault("HEALTHCHECK_ABOUT_LINK", None)
        app.config.setdefault("HEALTHCHECK_VIEW_NAME", "healthcheck")
        app.config.setdefault("HEALTHCHECK_PATH", "/healthcheck")
        app.config.setdefault("HEALTHCHECK_PARAM_KEY", "checks")
        app.config.setdefault("HEALTHCHECK_PARAM_SEP", "+")
        app.config.setdefault("HEALTHCHECK_CONTENT_TYPE", "application/health+json")

    def register_route(self, app, decorators=()):
        """

        :param app: Flask app or blueprint instance
        :param decorators: functions that decorates the healthcheck view
        """
        view = self.perform
        for decorator in decorators:
            view = decorator(view)

        conf = cap.config
        app.add_url_rule(
            conf["HEALTHCHECK_PATH"], conf["HEALTHCHECK_VIEW_NAME"], view_func=view
        )

    def register_extensions(self, app, extensions):
        """

        :param app: Flask app instance
        :param extensions: functions and arguments that execute the checks
        """
        for ex in extensions:
            try:
                func = ex.pop("func")
                self.register(**ex)(func)
            except Exception as exc:  # pylint: disable=W0703
                app.logger.exception(exc)
                app.logger.error("invalid healthcheck extension: %s", ex)

    @builder.response("json")
    def perform(self):
        healthy = True
        response = dict(
            status="pass",
            checks={},
            links={"about": cap.config["HEALTHCHECK_ABOUT_LINK"]},
        )

        params = flask.request.args.get(cap.config["HEALTHCHECK_PARAM_KEY"])
        all_checks = set(self._health_checks.keys())
        if not params:
            params = all_checks
        else:
            params = params.split(cap.config["HEALTHCHECK_PARAM_SEP"])
            params = set(params).intersection(all_checks)

        # noinspection PyProtectedMember,PyUnresolvedReferences
        app = cap._get_current_object()  # pylint: disable=W0212
        params = list(params)
        tasks = [(self._health_checks.get(p), dict(app=app)) for p in params]
        resp = self._executor(tasks=tasks).run()

        for i, p in enumerate(params):
            state, message = resp[i]
            status = "pass" if state else "fail"
            response["checks"][p] = dict(status=status, output=message)
            if not state:
                healthy = False

        if not healthy:
            response["status"] = "fail"

        status = httpcode.SUCCESS if healthy else httpcode.SERVICE_UNAVAILABLE
        return (
            response,
            status,
            {"Content-Type": cap.config["HEALTHCHECK_CONTENT_TYPE"]},
        )

    def register(self, name=None, **kwargs):  # pylint: disable=W0613
        """

        :param name:
        :return:
        """

        def _register(func):
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                self._health_checks[name] = functools.partial(func, **kwargs)

            return wrapped()

        return _register


health_checks = HealthCheck()



from .batch import FlaskelHTTPBatch, HTTPBatch
from .client import (
    all_errors,
    FlaskelHttp,
    FlaskelJsonRPC,
    HTTPBase,
    HTTPClient,
    HTTPStatusError,
    HTTPTokenAuth,
    NetworkError,
)
from .httpdumper import BaseHTTPDumper, FlaskelHTTPDumper, LazyHTTPDumper
from .useragent import UserAgent



import functools

from flask_cloudflare_remote import CloudflareRemote
from flask_cors import CORS
from flask_logify import FlaskLogging
from flask_response_builder import ResponseBuilder
from flask_template_support import TemplateSupport

cors = CORS()
logger = FlaskLogging()
builder = ResponseBuilder()
template = TemplateSupport()
cfremote = CloudflareRemote()

json_builder = functools.partial(builder.on_format("json"))



import os
import shutil
import sys
from pathlib import Path

import click

from flaskel.ext.sqlalchemy.schema import db_to_schema
from flaskel.ext.sqlalchemy.schema import model_to_uml
from flaskel.tester import cli as cli_tester


@click.group()
def cli():
    pass


@cli.command(name="init")
@click.argument("name")
def init(name):
    """Create skeleton for new application"""
    from flaskel import scripts as flaskel_scripts  # pylint: disable=C0415

    try:
        source = os.path.join(flaskel_scripts.__path__[0], "skeleton")
        shutil.copytree(source, ".", dirs_exist_ok=True)
        shutil.move("skel", name)
    except OSError as e:
        print(f"Unable to create new app. Error: {e}", file=sys.stderr)
        sys.exit(1)

    def replace_in_file(file, *args):
        _f = Path(file)
        text = _f.read_text()
        for sd in args:
            text = text.replace(*sd)
        _f.write_text(text)

    init_file = Path(os.path.join(name, "__init__.py"))
    init_file.write_text("from .version import *\n")

    for f in (
        "setup.py",
        "Dockerfile",
        "Makefile",
        "pytest.ini",
        ".coveragerc",
        ".bumpversion.cfg",
    ):
        replace_in_file(f, ("{skeleton}", name))

    replace_in_file(
        os.path.join(name, "scripts", "cli.py"),
        ("from ext", f"from {name}.ext"),
        ("from blueprint", f"from {name}.blueprint"),
    )
    replace_in_file(
        os.path.join(name, "scripts", "gunicorn.py"),
        ("from . import config", f"from {name}.scripts import config"),
    )
    replace_in_file(
        os.path.join("tests", "__init__.py"),
        ("from scripts.cli", f"from {name}.scripts.cli"),
    )


@cli.command(name="tests")
def run_tests():
    """Run test suite from package tests in current directory"""
    cli_tester.main()


@cli.command(name="schema")
@click.option("-m", "--from-models", help="module of sqlalchemy models")
@click.option("-d", "--from-database", help="database url connection")
@click.option("-f", "--file", required=True, help="output png filename")
def dump_schema(from_models, from_database, file):
    """Create png schema of database"""
    if from_models:
        graph = model_to_uml(from_models)
    elif from_database:
        graph = db_to_schema(from_database)
    else:
        raise click.UsageError("One of -m or -d are required")

    try:
        graph.write_png(file)  # pylint: disable=E1101
    except OSError as exc:
        print(str(exc), file=sys.stderr)
        print("try to install 'graphviz'", file=sys.stderr)


if __name__ == "__main__":
    cli()



from .batch import (
    AsyncBatchExecutor,
    BatchExecutor,
    DaemonThread,
    Thread,
    ThreadBatchExecutor,
)
from .datastruct import ConfigProxy, ExtProxy, HashableDict, IntEnum, ObjectDict
from .datetime import Day, from_iso_format, Millis, Minutes, Seconds, to_iso_format
from .schemas import JSONSchema, PayloadValidator, SCHEMAS



try:  # pragma: no cover
    import holidays
except ImportError:  # pragma: no cover
    holidays = None

import typing as t
from datetime import datetime, date as datetime_date

from dateutil.parser import parse as date_parse


class Millis:
    seconds = 1000
    minute = seconds * 60
    hour = minute * 60
    day = hour * 24


class Seconds:
    millis = 1000
    minute = 60
    hour = minute * 60
    day = hour * 24


class Minutes:
    seconds = 60
    hour = 60
    day = hour * 24


class Day:
    hours = 24
    minutes = hours * 60
    seconds = minutes * 60


DateType = t.Union[datetime_date, datetime]
AnyDateType = t.Union[DateType, str]


class DateHelper:
    DATE_PRETTY_FORMAT = "%d %B %Y %I:%M %p"
    DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"

    def __init__(self, *args, **kwargs):
        assert holidays is not None, "you must install holidays"
        self._holidays = holidays.CountryHoliday(*args, **kwargs)

    @property
    def holidays(self):
        return self._holidays  # pragma: no cover

    def all_holidays(self) -> list:
        return list(self._holidays.items())

    @classmethod
    def country_holidays(cls, *args, **kwargs):
        return holidays.CountryHoliday(*args, **kwargs)

    @classmethod
    def is_weekend(cls, curr_date: AnyDateType, **kwargs) -> bool:
        """

        :param curr_date: date instance or string date
        :param kwargs: passed to dateutil.parser.parse
        :return: boolean, True if is weekend
        """
        if isinstance(curr_date, str):
            curr_date = date_parse(curr_date, **kwargs)

        return curr_date.weekday() > 4

    @classmethod
    def change_format(
        cls,
        str_date: str,
        out_fmt: str,
        in_fmt: t.Optional[str] = None,
        raise_exc: bool = True,
    ) -> t.Optional[str]:
        """

        :param str_date: input string date
        :param out_fmt: format output date
        :param in_fmt: format input date (optional: could be detected from string)
        :param raise_exc: raise or not exception (default True)
        :return: return formatted date
        """
        if not (raise_exc or str_date):
            return None  # pragma: no cover

        try:
            date_time = cls.str_to_date(str_date, in_fmt)
            return date_time.strftime(out_fmt)
        except (ValueError, TypeError):
            if raise_exc is True:
                raise  # pragma: no cover
        return None

    @classmethod
    def now_iso_string(cls, is_utc: bool = True, tz=None) -> str:
        now = datetime.utcnow() if is_utc else datetime.now()
        return cls.date_to_str(now, tz)

    @classmethod
    def date_to_str(cls, date: DateType, fmt: t.Optional[str] = None) -> str:
        return date.strftime(fmt or cls.DATE_ISO_FORMAT)

    @classmethod
    def str_to_date(
        cls, date: str, fmt: t.Optional[str] = None, is_iso: bool = False, **kwargs
    ) -> datetime:
        if is_iso is True:
            fmt = cls.DATE_ISO_FORMAT
        if fmt is not None:
            return datetime.strptime(date, fmt)
        return date_parse(date, **kwargs)

    @classmethod
    def pretty_date(cls, date: AnyDateType):
        if isinstance(date, str):
            return cls.change_format(
                date, out_fmt=cls.DATE_PRETTY_FORMAT, raise_exc=True
            )
        return cls.date_to_str(date, fmt=cls.DATE_PRETTY_FORMAT)


# for backwards compatibility
def from_iso_format(str_date, fmt, exc=True):
    return DateHelper.change_format(
        str_date, in_fmt=DateHelper.DATE_ISO_FORMAT, out_fmt=fmt, raise_exc=exc
    )


# for backwards compatibility
def to_iso_format(str_date, fmt=None, exc=True):
    return DateHelper.change_format(
        str_date, in_fmt=fmt, out_fmt=DateHelper.DATE_ISO_FORMAT, raise_exc=exc
    )



from flaskel import ObjectDict

# noinspection PyUnresolvedReferences
from flaskel.tester import helpers as h  # noqa

VIEWS = ObjectDict(
    access_token="auth.access_token",
    refresh_token="auth.refresh_token",
    revoke_token="auth.revoke_token",
)



# example auth blueprint
#
from flask import Blueprint

from flaskel.ext import cors, error_handler

bp_auth = Blueprint("auth", __name__, subdomain="api", url_prefix="/auth")

cors.init_app(bp_auth)
error_handler.api_register(bp_auth)



import flask

from flaskel import cap, httpcode, ObjectDict

try:
    import stripe
    from stripe.error import StripeError
except ImportError:
    stripe = StripeError = None


class PaymentHandler:
    payment_intent_ok = "payment_intent.succeeded"
    payment_intent_ko = "payment_intent.payment_failed"

    def __init__(self, app=None, **kwargs):
        self._stripe = stripe
        self._success = self._not_registered
        self._error = self._not_registered

        if app is not None:
            self.init_app(app, **kwargs)

    @staticmethod
    def _not_registered(*_, **__):
        cap.logger.warning("no callback registered")

    def init_app(self, app, name=None, **kwargs):
        assert self._stripe is not None, "you must install 'stripe'"
        assert app.config.STRIPE_SECRET_KEY is not None, "missing STRIPE_SECRET_KEY"
        assert app.config.STRIPE_PUBLIC_KEY is not None, "missing STRIPE_PUBLIC_KEY"

        app.config.setdefault("STRIPE_DEBUG", False)
        app.config.setdefault("STRIPE_DEFAULT_CURRENCY", "eur")
        app.config.setdefault("STRIPE_API_VERSION", "2020-08-27")

        self._stripe.api_key = app.config.STRIPE_SECRET_KEY
        self._stripe.api_version = app.config.STRIPE_API_VERSION
        self._stripe.log = "debug" if app.config.STRIPE_DEBUG else "info"
        self._stripe.set_app_info(name or app.config.APP_NAME, **kwargs)

        if not hasattr(app, "extensions"):
            app.extensions = dict()  # pragma: no cover
        app.extensions["stripe"] = self

    def create_payment_intent(self, amount, currency=None, **kwargs):
        return ObjectDict(
            **self._stripe.PaymentIntent.create(
                amount=amount,
                currency=currency or cap.config.STRIPE_DEFAULT_CURRENCY,
                **kwargs,
            )
        )

    def create_event(self, data, raise_exc=False):
        webhook_secret = cap.config.STRIPE_WEBHOOK_SECRET
        signature = flask.request.headers.get("stripe-signature")
        try:
            return ObjectDict(
                **self._stripe.Webhook.construct_event(
                    payload=data, sig_header=signature, secret=webhook_secret
                )
            )
        except StripeError as exc:
            cap.logger.exception(exc)
            if raise_exc:
                raise
        return None

    def webhook_handler(self):
        data = self.create_event(flask.request.data)
        event_type = data.get("type")
        if event_type == self.payment_intent_ok:
            return self._success(data)
        if event_type == self.payment_intent_ko:
            return self._error(data)

        cap.logger.warning("unhandled event type received: %s\n%s", event_type, data)
        flask.abort(httpcode.NOT_FOUND)
        return None

    def on_success(self, callback):
        self._success = callback
        return callback

    def on_error(self, callback):
        self._error = callback
        return callback


payment_handler = PaymentHandler()



import os

import pytest
from scripts.cli import APP_CONFIG

from flaskel import ObjectDict, TestClient
from flaskel.ext.sqlalchemy.support import SQLASupport
from . import helpers as h

__all__ = [
    "h",
    "testapp",
    "test_client",
    "auth_token",
]

DB_TEST = "test.sqlite"
BASE_DIR = os.path.abspath(os.path.dirname(__file__))
SAMPLE_DATA = os.path.join(BASE_DIR, "..", "res", "sample.sql")
CONFIG = ObjectDict()

CONFIG.MAIL_DEBUG = True
CONFIG.CACHE_TYPE = "null"
CONFIG.RATELIMIT_ENABLED = False
CONFIG.SQLALCHEMY_ECHO = True
CONFIG.SQLALCHEMY_DATABASE_URI = f"sqlite:///{DB_TEST}"

CONFIG.SENDRIA = dict(
    endpoint="http://sendria.local:61000/api/messages",
    username="guest123",
    password="guest123",
)


def load_sample_data():
    SQLASupport.exec_from_file(
        CONFIG.SQLALCHEMY_DATABASE_URI, SAMPLE_DATA, CONFIG.SQLALCHEMY_ECHO
    )


EXTRAS = dict(
    conf=CONFIG,
    template_folder="blueprints/web/templates",
    static_folder="blueprints/web/static",
    callback=load_sample_data,
)


@pytest.fixture(scope="session")
def testapp():
    try:
        os.remove(DB_TEST)
    except OSError as exc:
        print(str(exc))

    return TestClient.get_app(**{**APP_CONFIG, **EXTRAS})


@pytest.fixture(scope="session")
def test_client(testapp):
    return testapp.test_client()


@pytest.fixture()
def auth_token(test_client):
    def get_access_token(token=None, email=None, password=None, in_query=True):
        if token is not None:
            if in_query is True:
                return f"jwt={token}"
            return dict(Authorization=f"Bearer {token}")

        credentials = dict(
            email=email or h.config.ADMIN_EMAIL,
            password=password or h.config.ADMIN_PASSWORD,
        )
        tokens = test_client.post(h.url_for(h.VIEWS.access_token), json=credentials)
        if in_query is True:
            return f"jwt={tokens.json.access_token}"
        return dict(Authorization=f"Bearer {tokens.json.access_token}")

    return get_access_token



__version__ = "0.0.0"

__author_name__ = "cs91chris"
__author_email__ = "cs91chris@voidbrain.me"
__url_version__ = "/v" + __version__.split(".")[0]

__author__ = f"{__author_name__} <{__author_email__}>"

__all__ = [
    "__version__",
    "__author__",
    "__author_name__",
    "__author_email__",
    "__url_version__",
]



import flask
from flask.views import MethodView, View

from flaskel.ext.default import builder
from flaskel.flaskel import httpcode


class BaseView(View):
    methods = ["GET"]

    def dispatch_request(self, *_, **__):
        """
        Must be implemented in every subclass
        """
        return self._not_implemented()  # pragma: no cover

    # noinspection PyMethodMayBeStatic
    def _not_implemented(self):  # pragma: no cover
        flask.abort(httpcode.NOT_IMPLEMENTED)

    @classmethod
    def register(cls, app, name, urls=(), **kwargs):
        """

        :param app: Flask app or blueprint
        :param name: optional view name
        :param urls: optional urls or view name
        :param kwargs: argument passed to cls constructor
        """
        name = name or cls.__name__
        view = cls.as_view(name, **kwargs)
        for u in urls or (f"/{name}",):
            app.add_url_rule(u, view_func=view, methods=cls.methods)


class Resource(MethodView):
    methods_collection = ["GET", "POST"]
    methods_subresource = ["GET", "POST"]
    methods_resource = ["GET", "PUT", "DELETE"]

    @staticmethod
    def normalize_url(url):
        return f"/{url.lstrip('/')}"

    @classmethod
    def register(cls, app, name=None, url=None, view=None, pk_type="int", **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name: view name
        :param url: url to bind if missing, name is used
        :param view: view class, subclass che call this must pass its reference
        :param pk_type: type of res_id
        """
        _class = view or cls
        name = name or _class.__name__
        url = cls.normalize_url(url or name)
        view_func = _class.as_view(name, **kwargs)

        if cls.methods_collection:
            app.add_url_rule(url, view_func=view_func, methods=cls.methods_collection)
        if cls.methods_resource:
            app.add_url_rule(
                f"{url}/<{pk_type}:res_id>",
                view_func=view_func,
                methods=cls.methods_resource,
            )
        if cls.methods_subresource:
            app.add_url_rule(
                f"{url}/<{pk_type}:res_id>/<sub_resource>",
                view_func=view_func,
                methods=cls.methods_subresource,
            )
        return view_func

    @builder.no_content
    def head(self, *args, **kwargs):
        return self.get(*args, **kwargs)

    @builder.on_accept()
    def get(self, *args, res_id=None, sub_resource=None, **kwargs):
        if res_id is None:
            return self.on_collection(*args, **kwargs)
        if sub_resource is None:
            return self.on_get(res_id, *args, **kwargs)
        if self.methods_subresource and "GET" not in self.methods_subresource:
            flask.abort(httpcode.METHOD_NOT_ALLOWED)

        _sub_resource = getattr(self, f"sub_{sub_resource}", None)
        if _sub_resource is None:
            flask.abort(httpcode.NOT_FOUND)
        return _sub_resource(res_id, *args, **kwargs)

    @builder.on_accept()
    def post(self, *args, res_id=None, sub_resource=None, **kwargs):
        if res_id is None:
            return self.on_post(*args, **kwargs)
        if self.methods_subresource and "POST" not in self.methods_subresource:
            flask.abort(httpcode.METHOD_NOT_ALLOWED)

        _sub_resource = getattr(self, f"sub_{sub_resource}_post", None)
        if _sub_resource is None:
            flask.abort(httpcode.NOT_FOUND)
        return _sub_resource(res_id, *args, **kwargs)

    @builder.no_content
    def delete(self, res_id, *args, **kwargs):
        return self.on_delete(res_id, *args, **kwargs)

    @builder.on_accept()
    def put(self, res_id, *args, **kwargs):
        return self.on_put(res_id, *args, **kwargs)

    def on_get(self, *_, **__):
        return self._not_implemented()  # pragma: no cover

    def on_post(self, *_, **__):
        return self._not_implemented()  # pragma: no cover

    def on_put(self, *_, **__):
        return self._not_implemented()  # pragma: no cover

    def on_delete(self, *_, **__):
        return self._not_implemented()  # pragma: no cover

    def on_collection(self, *_, **__):
        return self._not_implemented()  # pragma: no cover

    # noinspection PyMethodMayBeStatic
    def _not_implemented(self):  # pragma: no cover
        flask.abort(httpcode.NOT_IMPLEMENTED)



# based on https://github.com/enricobarzetti/sqlalchemy_get_or_create
from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound as NoResultError
from sqlalchemy.sql import text as execute_sql


class SQLASupport:
    def __init__(self, model, session):
        """

        :param model: a model class
        :param session: a session object
        """
        self.session = session
        self.model = model

    @staticmethod
    def _prepare_params(defaults=None, **kwargs):
        """

        :param defaults: overrides kwargs
        :param kwargs: overridden by defaults
        :return: merge of kwargs and defaults
        """
        ret = {}
        defaults = defaults or {}
        ret.update(kwargs)
        ret.update(defaults)
        return ret

    def _create_object(self, lookup, params, lock=False):
        """

        :param lookup: attributes used to find record
        :param params: attributes used to create record
        :param lock: flag used for atomic update
        :return:
        """
        obj = self.model(**params)
        self.session.add(obj)

        try:
            with self.session.begin_nested():
                self.session.flush()
        except IntegrityError:
            self.session.rollback()
            query = self.session.query(self.model).filter_by(**lookup)
            if lock:
                query = query.with_for_update()

            obj = query.one()
            return obj, False
        else:
            return obj, True

    def get_or_create(self, defaults=None, **kwargs):
        """

        :param defaults: attribute used to create record
        :param kwargs: filters used to fetch record or create
        :return:
        """
        try:
            return self.session.query(self.model).filter_by(**kwargs).one(), False
        except NoResultError:
            params = self._prepare_params(defaults, **kwargs)
            return self._create_object(kwargs, params)

    def update_or_create(self, defaults=None, **kwargs):
        """

        :param defaults: attribute used to create record
        :param kwargs: filters used to fetch record or create
        :return:
        """
        defaults = defaults or {}
        with self.session.begin_nested():
            try:
                query = self.session.query(self.model).with_for_update()
                obj = query.filter_by(**kwargs).one()
            except NoResultError:
                params = self._prepare_params(defaults, **kwargs)
                obj, created = self._create_object(kwargs, params, lock=True)
                if created:
                    return obj, created

            for k, v in defaults.items():
                setattr(obj, k, v)

            self.session.add(obj)
            self.session.flush()

        return obj, False

    @staticmethod
    def exec_from_file(db_uri, filename, echo=False):
        """

        :param db_uri:
        :param filename:
        :param echo:
        """
        engine = create_engine(db_uri, echo=echo)
        with engine.connect() as conn, open(filename) as f:
            for statement in f.read().split(";"):
                conn.execute(execute_sql(statement))



from flaskel.ext import (
    caching,
    client_mail,
    client_redis,
    date_helper,
    default,
    errors,
    limit,
    scheduler,
    useragent,
)
from flaskel.ext.auth import jwtm
from flaskel.ext.crypto import argon2
from flaskel.ext.healthcheck import health_checks
from flaskel.ext.sqlalchemy import db as sqlalchemy
from flaskel.extra.mobile_support import mobile_version, RedisStore
from flaskel.extra.stripe import payment_handler


class OPTS:
    cors = dict(resources={r"/*": {"origins": "*"}})
    mobile_version = dict(store=RedisStore(client_redis))
    errors = dict(
        dispatcher="subdomain",
        response=default.builder.on_accept(strict=False),
    )


EXTENSIONS = {
    # "name":   (<extension>, parameters: dict)
    "cfremote": (default.cfremote,),  # MUST be the first
    "logger": (default.logger,),  # MUST be the second
    "template": (default.template,),
    "builder": (default.builder,),
    "date_helper": (date_helper,),
    "cors": (default.cors, OPTS.cors),
    "database": (sqlalchemy,),
    "limiter": (limit.limiter,),
    "ip_ban": (limit.ip_ban,),
    "cache": (caching,),
    "errors": (errors.error_handler, OPTS.errors),
    "useragent": (useragent,),
    "argon2": (argon2,),
    "caching": (caching,),
    "scheduler": (scheduler,),
    "sendmail": (client_mail,),
    "redis": (client_redis,),
    "health_checks": (health_checks,),
    "jwt": (jwtm,),
    "ipban": (limit.ip_ban,),
    "stripe": (payment_handler,),
    "mobile_version": (mobile_version, OPTS.mobile_version),
}



import functools
import inspect
import typing as t

import flask
from flask.views import View

from flaskel.ext.default import builder
from flaskel.flaskel import Response
from flaskel.flaskel import cap
from flaskel.flaskel import httpcode
from flaskel.http import rpc
from flaskel.utils.batch import BatchExecutor
from flaskel.utils.datastruct import ObjectDict

RPCPayloadType = t.Union[t.List[dict], dict]
OperationsType = t.Dict[t.Optional[str], t.Dict[str, t.Callable]]


class JSONRPCView(View):
    version = "2.0"
    separator = "."
    operations: OperationsType = {}
    default_view_name = "jsonrpc"
    default_url = "/jsonrpc"
    methods = ["POST"]

    decorators = [
        builder.response("json"),
    ]

    @staticmethod
    def normalize_url(url):
        return f"/{url.lstrip('/')}"

    def __init__(
        self,
        operations: t.Optional[dict] = None,
        batch_executor: t.Type[BatchExecutor] = BatchExecutor,
        **kwargs,
    ):
        """

        :param batch_executor
        :param operations: a dict of jsonrpc operations like:
            {
                None: {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                "method1": {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                ...
            }
        """
        self.operations = operations or {}
        self._batch_executor = batch_executor
        self._batch_args = kwargs

    def _validate_request(self, data: dict):
        if "jsonrpc" not in data or "method" not in data:
            raise rpc.RPCInvalidRequest() from None

        if data["jsonrpc"] != self.version:
            raise rpc.RPCInvalidRequest(
                f"jsonrpc version is {self.version}", req_id=data.get("id")
            ) from None

    def _validate_payload(self) -> RPCPayloadType:
        payload: RPCPayloadType = flask.request.get_json(True)
        if not payload:
            raise rpc.RPCParseError() from None

        if isinstance(payload, (list, tuple)):
            max_requests = cap.config.JSONRPC_BATCH_MAX_REQUEST  # type: ignore
            if max_requests and len(payload) > max_requests:
                mess = f"Operations in a single http request must be less than {max_requests}"
                flask.abort(httpcode.REQUEST_ENTITY_TOO_LARGE, mess)
            for d in payload:
                self._validate_request(d)
        else:
            self._validate_request(payload)

        return payload

    def _get_action(self, method: str):
        """

        :param method:
        :return:
        """
        m = method.split(self.separator)
        if len(m) > 1:
            op, action = m[0], self.separator.join(m[1:])
        else:
            op, action = None, m[0]

        try:
            return self.operations[op][action]
        except (IndexError, TypeError, KeyError) as exc:
            cap.logger.debug(exc)
            raise rpc.RPCMethodNotFound()

    def dispatch_request(self):
        """

        :return:
        """
        tasks = []
        responses = []

        try:
            payload = self._validate_payload()
        except rpc.RPCError as ex:
            return (
                ObjectDict(
                    jsonrpc=self.version,
                    id=getattr(ex, "req_id", None),
                    error=ex.as_dict(),
                ),
                httpcode.BAD_REQUEST,
            )

        for d in payload if isinstance(payload, list) else [payload]:
            resp = ObjectDict(jsonrpc=self.version, id=None)
            try:
                if "id" not in d:
                    tasks.append(
                        (self._get_action(d["method"]), {**(d.get("params") or {})})
                    )
                else:
                    resp.id = d.get("id")
                    action = self._get_action(d["method"])
                    resp.result = action(**(d.get("params") or {}))
            except rpc.RPCError as ex:
                resp.error = ex.as_dict()
            except Exception as ex:  # pylint: disable=W0703
                cap.logger.exception(ex)
                mess = str(ex) if cap.debug is True else None
                resp.error = rpc.RPCInternalError(message=mess).as_dict()

            if "id" in d:
                responses.append(resp)

        self._batch_executor(tasks=tasks, **self._batch_args).run()

        if not responses:
            res = Response.no_content()
            return None, res.status_code, res.headers

        if isinstance(payload, (list, tuple)):
            if len(responses) > 1:
                return responses, httpcode.MULTI_STATUS
            return responses

        return responses[0]

    @classmethod
    def load_from_object(cls, obj):
        """

        :param obj:
        """
        for name, member in inspect.getmembers(obj, predicate=inspect.isroutine):
            if not name.startswith("_"):
                cls.method(obj.__class__.__name__, name)(member)

    @classmethod
    def method(cls, name=None, operation=None):  # pylint: disable=W0613
        """

        :param name:
        :param operation:
        :return:
        """

        def _method(func):
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                obj = {operation: func}
                if name not in cls.operations:
                    cls.operations[name] = obj
                else:
                    cls.operations[name].update(obj)

            return wrapped()

        return _method

    @classmethod
    def register(cls, app, name=None, url=None, **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name: view name
        :param url: url to bind if missing, name is used
        """
        name = name or cls.__name__
        url = cls.normalize_url(url or name)
        view_func = cls.as_view(name, cls.operations, **kwargs)
        app.add_url_rule(url, view_func=view_func, methods=cls.methods)



from base64 import b64encode
from functools import partial

import flask

from flaskel import ConfigProxy
from flaskel import httpcode
from flaskel.ext.sqlalchemy.support import SQLASupport
from flaskel.tester import FetchMail
from flaskel.tester.mixins import Asserter

__all__ = [
    "ConfigProxy",
    "httpcode",
    "Asserter",
    "config",
    "schemas",
    "url_for",
    "basic_auth_header",
    "restful_tester",
    "api_tester",
    "fetch_emails",
    "load_sample_data",
]

config = ConfigProxy()
schemas = ConfigProxy("SCHEMAS")
url_for = partial(flask.url_for, _external=True)


def load_sample_data(filename):
    SQLASupport.exec_from_file(
        config.SQLALCHEMY_DATABASE_URI, filename, echo=config.SQLALCHEMY_ECHO
    )


def fetch_emails(subject, recipient=None):
    client = FetchMail(config.SENDRIA)
    return client.perform(recipient=recipient, subject=subject)


def basic_auth_header(username=None, password=None):
    username = username or config.BASIC_AUTH_USERNAME
    password = password or config.BASIC_AUTH_PASSWORD
    token = b64encode(f"{username}:{password}".encode()).decode()
    return dict(Authorization=f"Basic {token}")


def api_tester(
    client, url=None, view=None, schema=None, status=httpcode.SUCCESS, **kwargs
):
    kwargs.setdefault("method", "GET")
    res = client.open(url_for(view) if view else url, **kwargs)
    Asserter.assert_status_code(res, status)
    if schema:
        Asserter.assert_schema(res.json, schema)
    return res


def restful_tester(
    client,
    view,
    headers=None,
    res_id=None,
    body_create=None,
    body_update=None,
    schema_read=None,
    schema_collection=None,
    methods=("GET", "POST", "PUT", "DELETE"),
    **params,
):  # pylint: disable=R0913
    args = "&".join([f"{k}={v}" for k, v in params.items()])
    url_collection = f"{url_for(view)}?{args}"

    if "POST" in methods:
        res = client.post(url_collection, json=body_create, headers=headers)
        Asserter.assert_status_code(res, httpcode.CREATED)
        Asserter.assert_schema(res.json, schema_read)
        res_id = res.json.id

    url_resource = f"{url_for(view, res_id=res_id)}?{args}"
    if "GET" in methods:
        res = client.get(url_collection, headers=headers)
        Asserter.assert_status_code(res)
        Asserter.assert_schema(res.json, schema_collection)

        res = client.get(url_resource, headers=headers)
        Asserter.assert_status_code(res)
        Asserter.assert_schema(res.json, schema_read)

    if "PUT" in methods:
        body_update = body_update or body_create
        res = client.put(url_resource, headers=headers, json=body_update)
        Asserter.assert_status_code(res)
        Asserter.assert_schema(res.json, schema_read)

    if "DELETE" in methods:
        res = client.delete(url_resource, headers=headers)
        Asserter.assert_status_code(res, httpcode.NO_CONTENT)

        res = client.get(url_resource, headers=headers)
        Asserter.assert_status_code(res, httpcode.NOT_FOUND)
        Asserter.assert_schema(res.json, schemas.API_PROBLEM)



from multiprocessing import cpu_count

try:
    from waitress import serve
except ImportError:
    serve = None  # type: ignore

from .base import BaseApplication


class WSGIWaitress(BaseApplication):
    def run(self):
        """

        :return:
        """
        self.options.setdefault("threads", cpu_count())
        assert serve is not None, "You must install 'waitress'"

        serve(
            self.application,
            host=self._interface,
            port=self._port,
            threads=self.options["threads"],
        )



import os

from decouple import Choices, config, Csv

from flaskel.utils import logger, yaml, Seconds

DEBUG = config("DEBUG", default=False, cast=bool)
TESTING = config("TESTING", default=DEBUG, cast=bool)
APP_NAME = config("APP_NAME", default="flaskel")
APP_HOST = config("APP_HOST", default="127.0.0.1")
APP_PORT = config("APP_PORT", default=5000, cast=int)
FLASK_APP = config("FLASK_APP", default="app:app")
SERVER_NAME = config("SERVER_NAME", default=f"{APP_HOST}:{APP_PORT}")

FLASK_ENV = config(
    "FLASK_ENV",
    default="development" if DEBUG else "production",
    cast=Choices(["development", "production"]),
)

LOCALE = config("LOCALE", default="en_EN.utf8")
TEMPLATES_AUTO_RELOAD = config("TEMPLATES_AUTO_RELOAD", default=DEBUG, cast=bool)
EXPLAIN_TEMPLATE_LOADING = config("EXPLAIN_TEMPLATE_LOADING", default=False, cast=bool)
APIDOCS_ENABLED = config("APIDOCS_ENABLED", default=True, cast=bool)
CONF_PATH = config(
    "CONF_PATH", default=os.path.join("flaskel", "scripts", "skeleton", "res")
)

JWT_TOKEN_LOCATION = ["headers", "query_string"]
JWT_ACCESS_TOKEN_EXPIRES = config(
    "JWT_ACCESS_TOKEN_EXPIRES", default=Seconds.day, cast=int
)
JWT_REFRESH_TOKEN_EXPIRES = config(
    "JWT_REFRESH_TOKEN_EXPIRES", default=Seconds.day * 14, cast=int
)

SQLALCHEMY_DATABASE_URI = config("DATABASE_URL", default="sqlite:///db.sqlite")
MONGO_URI = config("MONGO_URI", default="mongodb://localhost:27017")
REDIS_URL = config("REDIS_URL", default="redis://127.0.0.1:6379")
REDIS_OPTS = {
    "decode_responses": True,
    "socket_connect_timeout": config("REDIS_CONN_TIMEOUT", default=0.05, cast=float),
}

BASIC_AUTH_USERNAME = config("BASIC_AUTH_USERNAME", default="admin")
BASIC_AUTH_PASSWORD = config("BASIC_AUTH_PASSWORD", default="admin")

MAIL_DEBUG = config("MAIL_DEBUG", default=DEBUG, cast=bool)
MAIL_SERVER = config("MAIL_SERVER", default="sendria.local")
MAIL_PORT = config("MAIL_SERVER", default=62000, cast=int)
ADMIN_EMAIL = config("ADMIN_EMAIL", default="admin@mail.com")
ADMIN_PASSWORD = config("ADMIN_PASSWORD", default="admin")
MAIL_DEFAULT_SENDER = config("MAIL_DEFAULT_SENDER", default="admin@mail.com")
MAIL_DEFAULT_RECEIVER = config("MAIL_DEFAULT_RECEIVER", default="admin@mail.com")

PREFERRED_URL_SCHEME = config(
    "PREFERRED_URL_SCHEME", default="http" if FLASK_ENV == "development" else "https"
)

IPBAN_COUNT = config("IPBAN_COUNT", default=5, cast=int)
IPBAN_SECONDS = config("IPBAN_SECONDS", default=Seconds.hour, cast=int)

LOG_BUILDER = config("LOG_BUILDER", default="text")
LOG_APP_NAME = config("LOG_APP_NAME", default=APP_NAME)
LOG_LOGGER_NAME = config("LOG_LOGGER_NAME", default=FLASK_ENV)
LOG_REQ_HEADERS = config("LOG_REQ_HEADERS", default="", cast=Csv())
LOG_RESP_HEADERS = config("LOG_RESP_HEADERS", default="", cast=Csv())
LOG_REQ_SKIP_DUMP = config("LOG_REQ_SKIP_DUMP", default=not TESTING, cast=bool)
LOG_RESP_SKIP_DUMP = config("LOG_RESP_SKIP_DUMP", default=not TESTING, cast=bool)
LOG_REQ_FORMAT = config(
    "LOG_REQ_FORMAT",
    default="INCOMING REQUEST {address} {method} {scheme} {path}{headers}{body}",
)
LOG_RESP_FORMAT = config(
    "LOG_RESP_FORMAT",
    default="OUTGOING RESPONSE for {address} at {path}: STATUS {status}{headers}{body}",
)

CF_STRICT_ACCESS = config("CF_STRICT_ACCESS", default=False, cast=bool)
VERSION_STORE_MAX = config("VERSION_STORE_MAX", default=6, cast=int)
VERSION_CACHE_EXPIRE = config("VERSION_CACHE_EXPIRE", default=Seconds.hour, cast=int)

HTTP_PROTECT_BODY = config("HTTP_PROTECT_BODY", default=False, cast=bool)
HTTP_DUMP_BODY = [
    config("HTTP_DUMP_REQ_BODY", default=False, cast=bool),
    config("HTTP_DUMP_RESP_BODY", default=False, cast=bool),
]

USE_X_SENDFILE = config("USE_X_SENDFILE", default=not DEBUG, cast=bool)
ENABLE_ACCEL = config("ENABLE_ACCEL", default=True, cast=bool)
ACCEL_BUFFERING = True
ACCEL_CHARSET = "utf-8"
ACCEL_LIMIT_RATE = "off"

WSGI_WERKZEUG_PROFILER_FILE = "profiler.txt"
WSGI_WERKZEUG_PROFILER_RESTRICTION = (0.1,)
WSGI_WERKZEUG_LINT_ENABLED = config(
    "WSGI_WERKZEUG_LINT_ENABLED", default=TESTING, cast=bool
)
WSGI_WERKZEUG_PROFILER_ENABLED = config(
    "WSGI_WERKZEUG_PROFILER_ENABLED", default=TESTING, cast=bool
)

SQLALCHEMY_TRACK_MODIFICATIONS = False
SQLALCHEMY_ECHO = config("SQLALCHEMY_ECHO", default=TESTING, cast=bool)

REQUEST_ID_HEADER = "X-Request-ID"
RATELIMIT_ENABLED = config("RATELIMIT_ENABLED", default=not DEBUG, cast=bool)
RATELIMIT_HEADERS_ENABLED = config("RATELIMIT_HEADERS_ENABLED", default=True, cast=bool)
RATELIMIT_IN_MEMORY_FALLBACK_ENABLED = config(
    "RATELIMIT_IN_MEMORY_FALLBACK_ENABLED", default=True, cast=bool
)

DATABASE_URL = SQLALCHEMY_DATABASE_URI
ERROR_PAGE = "core/error.html"
SECRET_KEY_MIN_LENGTH = 256
RB_DEFAULT_ACCEPTABLE_MIMETYPES = [
    "application/json",
    "application/xml",
]

JWT_DEFAULT_SCOPE = None
JWT_DEFAULT_TOKEN_TYPE = "bearer"
PRETTY_DATE = "%d %B %Y %I:%M %p"
DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"
SEND_FILE_MAX_AGE_DEFAULT = Seconds.day

RATELIMIT_STORAGE_URL = REDIS_URL
RATELIMIT_KEY_PREFIX = APP_NAME
RATELIMIT_STORAGE_OPTIONS = {
    "socket_timeout": REDIS_OPTS["socket_connect_timeout"],
    "socket_connect_timeout": REDIS_OPTS["socket_connect_timeout"],
}

LIMITER = {
    "FAIL": "1/second",
    "FAST": "30/minute",
    "MEDIUM": "20/minute",
    "SLOW": "10/minute",
    "BYPASS_KEY": "X-Limiter-Bypass",
    "BYPASS_VALUE": "bypass-rate-limit",
}

SCHEDULER_AUTO_START = config("SCHEDULER_AUTO_START", default=True, cast=bool)
SCHEDULER_API_ENABLED = config("SCHEDULER_API_ENABLED", default=False, cast=bool)
SCHEDULER_JOBSTORES = {
    "default": {
        "class": "apscheduler.jobstores.sqlalchemy:SQLAlchemyJobStore",
        "tablename": "scheduler",
        "url": DATABASE_URL,
    }
}
SCHEDULER_EXECUTORS = {"default": {"type": "threadpool", "max_workers": 20}}
SCHEDULER_JOB_DEFAULTS = {"coalesce": False, "max_instances": 10}

CACHE_REDIS_URL = REDIS_URL
CACHE_DEFAULT_TIMEOUT = Seconds.hour
CACHE_TYPE = "flask_caching.backends.redis"
CACHE_KEY_PREFIX = config("CACHE_KEY_PREFIX", default=APP_NAME)
CACHE_OPTIONS = {
    "socket_timeout": REDIS_OPTS["socket_connect_timeout"],
    "socket_connect_timeout": REDIS_OPTS["socket_connect_timeout"],
}

APISPEC = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, "swagger.yaml"))
SCHEMAS = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, "schemas.yaml"))
SCHEDULER_JOBS = yaml.load_optional_yaml_file(os.path.join(CONF_PATH, "scheduler.yaml"))
IPBAN_NUISANCES = yaml.load_optional_yaml_file(
    os.path.join(CONF_PATH, "nuisances.yaml")
)
LOGGING = yaml.load_optional_yaml_file(
    os.path.join(CONF_PATH, "log.yaml"), default=logger.LOGGING
)


__all__ = [s for s in dir() if s.isupper()]



# example api blueprint
#
from flask import Blueprint

from flaskel.ext import cors, error_handler

bp_api = Blueprint(
    "api", __name__, subdomain="api", static_folder="static", static_url_path="/cdn/"
)

cors.init_app(bp_api)
error_handler.api_register(bp_api)



import io
from functools import partial

import flask
import jsonschema
from flask import current_app as cap

from flaskel.http.client import HTTPClient
from flaskel.http.client import httpcode
from flaskel.utils.datastruct import ConfigProxy
from flaskel.utils.datastruct import ObjectDict

SCHEMAS = ConfigProxy("SCHEMAS")


class JSONSchema:
    loader = partial(flask.json.loads)
    dumper = partial(flask.json.dumps, indent=4)
    marker = "3fb539deef7c4e2991f265c0a982f5ea"
    message_format = "{message}\nError in line {line}:\n{report}\n{message}"

    @classmethod
    def load_from_url(cls, url):
        """

        :param url:
        :return:
        """
        res = HTTPClient(url, raise_on_exc=True).get(url)
        return res.body

    @classmethod
    def load_from_file(cls, file):
        """

        :param file:
        :return:
        """
        if file.startswith("file://"):
            file = file[7:]

        with open(file) as f:
            return cls.loader(f.read())

    @classmethod
    def validate(cls, data, schema, raise_exc=False, pretty_error=True, checker=None):
        """

        :param data:
        :param schema:
        :param raise_exc:
        :param pretty_error:
        :param checker:
        :return:
        """
        if isinstance(schema, str):
            if schema.startswith("https://") or schema.startswith("http://"):
                schema = cls.load_from_url(schema)
            if schema.startswith("file://"):
                schema = cls.load_from_file(schema)

        try:
            checker = checker or jsonschema.FormatChecker()
            jsonschema.validate(data, schema, format_checker=checker)
        except (jsonschema.ValidationError, jsonschema.SchemaError) as exc:
            if not raise_exc:
                if pretty_error:
                    cap.logger.error(cls.error_report(exc, data))
                else:
                    cap.logger.exception(exc)
                return False
            raise
        return True

    @classmethod
    def error_report(cls, e, json_object, lines_before=8, lines_after=8):
        """
        From: https://github.com/ccpgames/jsonschema-errorprinter/blob/master/jsonschemaerror.py

        Generate a detailed report of a schema validation error.
        'e' is a jsonschema.ValidationError exception raised on 'json_object'.

        Steps to discover the location of the validation error:
            1. Traverse the json object using the 'path' in the validation exception
               and replace the offending value with a special marker.
            2. Pretty-print the json object indented json text.
            3. Search for the special marker in the json text to find the actual
               line number of the error.
            4. Make a report by showing the error line with a context of
               'lines_before' and 'lines_after' number of lines on each side.
        """
        if not e.path:
            return e.message or str(e)

        # Find the object that is erroring, and replace it with the marker.
        for entry in list(e.path)[:-1]:
            json_object = json_object[entry]

        orig, json_object[e.path[-1]] = json_object[e.path[-1]], cls.marker

        # Pretty print the object and search for the marker.
        json_error = cls.dumper(json_object)
        err_line = None

        for lineno, text in enumerate(io.StringIO(json_error)):
            if cls.marker in text:
                err_line = lineno
                break

        if not err_line:
            return e.message or str(e)

        report = []
        json_object[e.path[-1]] = orig
        json_error = cls.dumper(json_object)

        for lineno, text in enumerate(io.StringIO(json_error)):
            line_text = "{:4}: {}".format(
                lineno + 1, ">" * 3 if lineno == err_line else " " * 3
            )
            report.append(line_text + text.rstrip("\n"))

        report = report[max(0, err_line - lines_before) : err_line + 1 + lines_after]
        return cls.message_format.format(
            line=err_line + 1, report="\n".join(report), message=e.message or str(e)
        )


class PayloadValidator:
    schemas = SCHEMAS
    validator = JSONSchema

    @classmethod
    def validate(cls, schema, strict=True):
        """

        :param schema:
        :param strict:
        :return:
        """
        if strict and schema is None:
            flask.abort(httpcode.INTERNAL_SERVER_ERROR, "empty schema")

        payload = flask.request.json
        try:
            schema = cls.schemas.get(schema) if isinstance(schema, str) else schema
            cls.validator.validate(payload, schema, raise_exc=True)
            return payload
        except jsonschema.SchemaError as exc:
            cap.logger.exception(exc)
            flask.abort(httpcode.INTERNAL_SERVER_ERROR)
        except jsonschema.ValidationError as exc:
            cap.logger.error(cls.validator.error_report(exc, payload))
            reason = dict(cause=exc.cause, message=exc.message, path=exc.path)
            flask.abort(httpcode.UNPROCESSABLE_ENTITY, response=dict(reason=reason))
        return None


class Fields:
    schema = ObjectDict(**{"$schema": "http://json-schema.org/draft-07/schema#"})

    null = ObjectDict(type="null")
    integer = ObjectDict(type="integer")
    string = ObjectDict(type="string")
    number = ObjectDict(type="number")
    boolean = ObjectDict(type="boolean")
    datetime = ObjectDict(type="string", format="date-time")
    any_object = ObjectDict(type="object", additionalProperties=True)
    any = ObjectDict(
        type=["integer", "string", "number", "boolean", "array", "object", "null"]
    )

    class Opt:
        integer = ObjectDict(type=["integer", "null"])
        string = ObjectDict(type=["string", "null"])
        number = ObjectDict(type=["number", "null"])
        boolean = ObjectDict(type=["boolean", "null"])

    @classmethod
    def oneof(cls, *args, **kwargs):
        return ObjectDict(oneOf=args if len(args) > 1 else (*args, cls.null), **kwargs)

    @classmethod
    def anyof(cls, *args, **kwargs):
        return ObjectDict(anyOf=args if len(args) > 1 else (*args, cls.null), **kwargs)

    @classmethod
    def ref(cls, path, **kwargs):
        return ObjectDict(**{"$ref": f"#{path}", **kwargs})

    @classmethod
    def enum(cls, *args, **kwargs):
        return {"enum": args, **kwargs}

    @classmethod
    def type(cls, *args, **kwargs):
        return {"type": args, **kwargs}

    @classmethod
    def object(
        cls,
        required=(),
        not_required=(),
        properties=None,
        all_required=True,
        additional=False,
        **kwargs,
    ):
        properties = properties or {}
        if not required and all_required is True:
            required = [i for i in properties.keys() if i not in not_required]

        return ObjectDict(
            type="object",
            additionalProperties=additional,
            required=required,
            properties=properties,
            **kwargs,
        )

    @classmethod
    def array(cls, items, min_items=0, **kwargs):
        return ObjectDict(type="array", minItems=min_items, items=items, **kwargs)

    @classmethod
    def array_object(cls, min_items=0, **kwargs):
        return ObjectDict(type="array", minItems=min_items, items=cls.object(**kwargs))



from functools import partial

from flask import request
from flask_caching import Cache as FlaskCache

from flaskel.http import httpcode

caching = FlaskCache()


class Cache:
    cache = caching

    cached = partial(
        cache.cached,
        query_string=True,
        unless=lambda: "no-store" in (request.headers.get("Cache-Control") or ""),
        response_filter=lambda r: httpcode.is_ok(r.status_code),
    )



from flaskel.utils.schemas import default

SCHEMAS = dict(
    PostRevokeToken=default.SCHEMAS.POST_REVOKE_TOKEN,
    PostAccessToken=default.SCHEMAS.POST_ACCESS_TOKEN,
    AccessToken=default.SCHEMAS.ACCESS_TOKEN,
    RefreshToken=default.SCHEMAS.REFRESH_TOKEN,
    ApiProblem=default.SCHEMAS.API_PROBLEM,
)

if __name__ == "__main__":  # pragma: no cover
    import json

    print(json.dumps(SCHEMAS))



# example web blueprint
#
from flask import Blueprint

from flaskel.ext import error_handler
from flaskel.views import RenderTemplate

bp_web = Blueprint(
    "web",
    __name__,
    url_prefix="/",
    template_folder="templates",
    static_folder="static",
    static_url_path="static/",
)

error_handler.web_register(bp_web)
RenderTemplate.register(bp_web, name="index", urls=["/"], template="index.html")



from .caching import Cache, caching
from .datetime import date_helper
from .default import builder, cfremote, cors, json_builder, logger, template
from .errors import error_handler, ErrorNormalizer
from .jobs import APJobs, scheduler
from .limit import FlaskIPBan, ip_ban, limiter, RateLimit
from .redis import client_redis, FlaskRedis
from .sendmail import client_mail, ClientMail
from .useragent import UserAgent, useragent



from typing import TYPE_CHECKING

from flask_sqlalchemy import event
from sqlalchemy.ext.declarative import declared_attr

from flaskel import ObjectDict
from flaskel.ext.crypto import argon2
from flaskel.ext.sqlalchemy import db

if TYPE_CHECKING:
    hybrid_property = property  # pylint: disable=C0103
else:
    from sqlalchemy.ext.hybrid import hybrid_property


class StandardMixin:
    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    created_at = db.Column(db.DateTime, server_default=db.func.now())
    updated_at = db.Column(db.DateTime, onupdate=db.func.now())


class CatalogMixin:
    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """

    @declared_attr
    def __table_args__(self):
        return (db.UniqueConstraint("label", "type_id"),)

    id = db.Column(db.Integer, primary_key=True)
    label = db.Column(db.String(100))
    type_id = db.Column(db.Integer)
    description = db.Column(db.String(250))

    def __str__(self):
        return str(self.label)

    def to_dict(self, **kwargs):
        return ObjectDict(
            id=self.id,
            label=self.label,
            description=self.description,
            typeId=self.type_id,
            **kwargs,
        )


class CatalogXMixin(CatalogMixin):
    @declared_attr
    def __table_args__(self):
        return (db.UniqueConstraint("code", "type_id"),)

    code = db.Column(db.String(20))
    order_id = db.Column(db.Integer, index=True)

    order_by = order_id.asc()

    def __str__(self):
        return f"<{self.code} - {self.label}>"

    def to_dict(self, **kwargs):
        return super().to_dict(code=self.code, orderId=self.order_id, **kwargs)


class LoaderMixin:
    values = ()

    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """

    @classmethod
    def load_values(cls, *_, **__):
        for d in cls.values:
            db.session.add(cls(**d))
        db.session.commit()

    @classmethod
    def register_loader(cls):
        # noinspection PyUnresolvedReferences
        event.listen(cls.__table__, "after_create", cls.load_values)


class UserMixin(StandardMixin):
    _hasher = argon2
    _password = db.Column("password", db.String(128), nullable=False)

    email = db.Column(db.String(255), unique=True, nullable=False)

    @hybrid_property
    def password(self):
        return self._password

    @password.setter
    def password(self, password):
        self._password = self._hasher.generate_hash(password)

    def check_password(self, password):
        return self._hasher.verify_hash(self._password, password)

    def to_dict(self, **_):
        return ObjectDict(id=self.id, email=self.email)



try:
    from flask_pymongo import PyMongo
except (ModuleNotFoundError, ImportError):
    PyMongo = object

from flaskel import ObjectDict


class FlaskMongoDB(PyMongo):
    def init_app(self, app, *args, uri=None, ext_name="default", **kwargs):
        """

        :param app:
        :param uri:
        :param ext_name:
        :param args:
        :param kwargs:
        """
        assert PyMongo is not object, "you must install 'flask_pymongo'"

        app.config.setdefault("MONGO_OPTS", {})
        app.config["MONGO_OPTS"].setdefault("connectTimeoutMS", 1000)
        app.config["MONGO_OPTS"].setdefault("serverSelectionTimeoutMS", 1000)
        app.config["MONGO_OPTS"].update(**kwargs)

        # noinspection PyUnresolvedReferences
        super().init_app(app, uri, *args, **kwargs)
        app.extensions["mongo"] = ObjectDict()
        app.extensions["mongo"][ext_name] = self



# example spa blueprint
#
import os

from flask import Blueprint

from flaskel import cap
from flaskel.ext import error_handler

static_folder = os.environ.get("SPA_STATIC_FOLDER") or "webapp"
static_url_path = os.environ.get("SPA_STATIC_URL_PATH") or "/assets/"

bp_spa = Blueprint(
    "spa", __name__, static_folder=static_folder, static_url_path=static_url_path
)

error_handler.web_register(bp_spa)


@bp_spa.route("/", defaults=dict(path=""))
@bp_spa.route("/<path:path>")
def catch_all(path):
    """

    :param path:
    :return:
    """
    return cap.send_static_file("index.html")



import inspect
import os
import sys
import threading
import traceback
from multiprocessing import cpu_count

from decouple import config as AutoConfig

from . import config as AppConfigFile

host = AutoConfig("APP_HOST", default="127.0.0.1")
port = AutoConfig("APP_PORT", default="5000", cast=int)
# A string of the form: 'HOST', 'HOST:PORT', 'unix:PATH'
bind = AutoConfig("BIND", default=f"{host}:{port}")

pidfile = AutoConfig("PID_FILE", default=".gunicorn.pid")
# This requires that you install the setproctitle module
proc_name = AutoConfig("PROC_NAME", default=None)
app_config_file = inspect.getmodule(AppConfigFile)
app_config_file = os.path.abspath(app_config_file.__file__)  # type: ignore
app_config = AutoConfig("APP_CONFIG_FILE", default=app_config_file)

timeout = AutoConfig("TIMEOUT", default=30, cast=int)
backlog = AutoConfig("BACKLOG", default=2048, cast=int)
keepalive = AutoConfig("KEEPALIVE", default=3, cast=int)

nw = 1 + 2 * cpu_count()  # generally in the 2-4 x $(NUM_CORES) range
worker_class = AutoConfig("WORKER_CLASS", default="sync")
workers = AutoConfig("WORKERS", default=nw if worker_class == "sync" else 1, cast=int)
threads = AutoConfig(
    "THREADS", default=nw if worker_class == "gthread" else 1, cast=int
)

# For eventlet and gevent, limits the max number of simultaneous clients
# that a single process can handle
worker_connections = AutoConfig("WORKER_CONNECTIONS", default=1000, cast=int)

# Install a trace function that spews every line of Python that is executed when running the server
spew = False
daemon = False
umask = 0
user = None
group = None

raw_env = [
    f"APP_CONFIG_FILE={app_config}",
]

errorlog = "-"
accesslog = "-"
loglevel = "info"  # One of "debug", "info", "warning", "error", "critical"
access_log_format = '%(h)s %(l)s %(u)s %(t)s "%(r)s" %(s)s %(b)s "%(f)s" "%(a)s"'


def post_fork(server, worker):
    server.log.info("Worker spawned (pid: %s)", worker.pid)


def pre_fork(server, worker):
    server.log.info("pre fork (pid: %s)", worker.pid)


def after_fork(server, worker):
    server.log.info("after fork (pid: %s)", worker.pid)


def pre_exec(server):
    server.log.info("Forked child, re-executing.")


def when_ready(server):
    server.log.info("Server is ready. Spawning workers")


def worker_abort(worker):
    worker.log.info("worker received SIGABRT signal")


def worker_int(worker):
    worker.log.info("worker received INT or QUIT signal")

    class DumpTraceback:
        def __str__(self):
            code = []
            id2name = {th.ident: th.name for th in threading.enumerate()}
            # noinspection PyUnresolvedReferences, PyProtectedMember
            for threadId, stack in sys._current_frames().items():
                code.append(
                    "\n# Thread: %s(%d)" % (id2name.get(threadId, ""), threadId)
                )
                for filename, lineno, name, line in traceback.extract_stack(stack):
                    code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                    if line:
                        code.append("  %s" % (line.strip()))
            return "\n".join(code)

    worker.log.debug("%s", DumpTraceback())



try:
    from gevent.pywsgi import WSGIServer
except ImportError:
    WSGIServer = object

from .base import BaseApplication


class WSGIGevent(BaseApplication, WSGIServer):
    def __init__(self, app, options=None):
        """

        :param app:
        :param options:
        """
        assert WSGIServer is not object, "you must install 'gevent'"
        BaseApplication.__init__(self, app, options)
        opts = dict(
            spawn=options.get("spawn") or "default", backlog=options.get("backlog")
        )
        WSGIServer.__init__(self, self._bind, self.application, **opts)

    def run(self):
        self.serve_forever()
